import builtins
import numpy as np


def info(data: dict, quantile=80, add_marker="!"):
    steps = np.array(data["steps"])
    quantile_value = np.quantile(steps, quantile / 100)

    max_number_sign = 20
    steps_over_q80 = np.where(steps >= quantile_value, steps, np.nan)

    max_value = np.max(steps)
    normalized = (steps / max_value) * max_number_sign
    max_normalized = np.max(normalized)
    lines = []
    for i, value in enumerate(normalized):
        bar = "#" * int(value)
        space = " " * int(max_normalized - value)
        marker = add_marker if steps[i] >= quantile_value else "  "
        lines.append(f"{i: 2d} | {bar}{space} {marker} {steps[i]: .2f}")

    return {
        "quantile": quantile,
        "quantile_value": quantile_value,
        "steps_over_quantile": steps_over_q80,
        "graphic": "\n".join(lines),
    }


def print(data: dict, quantile=80, add_marker="!"):
    result = info(data, quantile, add_marker)
    builtins.print("==" * 20)
    builtins.print(data["name"])
    builtins.print("-" * 20)
    builtins.print(result["graphic"])
    builtins.print("-" * 20)
    builtins.print("quantile value: " + str(result["quantile_value"]))
    builtins.print("==" * 20)
    return result



import time


__default = "__default"
__values = {}


def tic(name: str = __default) -> None:
    """First time"""
    time_seg = time.perf_counter()
    __values[name] = [time_seg]


def tac(name: str = __default) -> float:

    time_seg = time.perf_counter()

    __values[name].append(time_seg)
    return __values[name][-1] - __values[name][-2]


def toc(name: str = __default, skip_toc=False) -> dict:
    if not skip_toc:
        tac(name)
    steps = []
    for index in range(1, len(__values[name])):
        steps.append(__values[name][index] - __values[name][index - 1])
    total = __values[name][-1] - __values[name][0]
    del __values[name]
    return {"name": name, "total": total, "steps": steps}


def get_active_timers():
    return ", ".join(__values.keys())



"""Testing subpackage"""

from pathlib import Path
import pytest


def run_tests():
    """run all available tests"""
    tests_path = Path(__file__).parent
    print(f"Running tests from path: {tests_path}")
    pytest.main(["-v", "-m", "not slow", "-p" "no:warnings", tests_path])



import pytest
from time import sleep
from tictactoc import tictactoc


def test_tictoc():
    time_to_wait = 1
    tictactoc.tic()
    sleep(time_to_wait)
    assert pytest.approx(time_to_wait, 0.001) == tictactoc.toc()["total"]


def test_tictactoc():
    tac = 0.5
    toc = 0.34
    tictactoc.tic()
    sleep(tac)
    tictactoc.tac()
    sleep(toc)
    assert pytest.approx(tac + toc, 0.001) == tictactoc.toc()["total"]


def test_tictactoc2():
    tac = 0.3
    tac2 = 0.4
    toc = 0.2
    tictactoc.tic()
    sleep(tac)
    tictactoc.tac()
    sleep(tac2)
    tictactoc.tac()
    sleep(toc)
    assert pytest.approx(tac + tac2 + toc, 0.001) == tictactoc.toc()["total"]


def test_tictactoc_name():
    tac = 0.3
    tac2 = 0.4
    toc = 0.2
    tictactoc.tic("test")
    sleep(tac)
    tictactoc.tac("test")
    sleep(tac2)
    tictactoc.tac("test")
    sleep(toc)
    assert tictactoc.toc("test")["total"] == pytest.approx(
        tac + tac2 + toc, 0.001
    )


def test_clear_noname():
    first = 0.3
    second = 0.8
    tictactoc.tic()
    sleep(first)
    assert tictactoc.toc()["total"] == pytest.approx(first, 0.01)
    tictactoc.tic()
    sleep(second)
    assert tictactoc.toc()["total"] == pytest.approx(second, 0.01)


def test_stats():
    tac = 0.5
    tac2 = 0.2
    tac3 = 0.6
    toc = 0.1
    tictactoc.tic()
    sleep(tac)
    tictactoc.tac()
    sleep(tac2)
    tictactoc.tac()
    sleep(tac3)
    tictactoc.tac()
    sleep(toc)
    result = tictactoc.toc()
    assert result["total"] == pytest.approx(tac + tac2 + tac3 + toc, 0.01)
    assert result["steps"] == [
        pytest.approx(tac, 0.01),
        pytest.approx(tac2, 0.01),
        pytest.approx(tac3, 0.01),
        pytest.approx(toc, 0.01),
    ]


def test_active_timers():
    tictactoc.tic()
    tictactoc.tic("second timer")
    tictactoc.tic("another timer")

    active_timers = tictactoc.get_active_timers()
    assert active_timers == "__default, second timer, another timer"

    tictactoc.toc("second timer")
    active_timers = tictactoc.get_active_timers()
    assert active_timers == "__default, another timer"


def test_skip_toc():
    tictactoc.tic("my loop")

    my_loop = [1, 2]

    for element in my_loop:
        sleep(0.1)
        tictactoc.tac("my loop")

    result = tictactoc.toc("my loop", skip_toc=True)
    assert len(result["steps"]) == 2


def test_values_in_toc():
    tictactoc.tic()
    result = tictactoc.toc()
    assert "name" in result
    assert "total" in result
    assert "steps" in result
    assert len(result["name"]) > 0
    assert result["total"] >= 0
    assert len(result["steps"]) >= 0
